#!/usr/bin/env python
# coding: utf-8

# In[4]:


import argparse
import os

import torch
from torch.optim import AdamW
from torch.utils.data import DataLoader
from peft import (
    get_peft_config,
    get_peft_model,
    get_peft_model_state_dict,
    set_peft_model_state_dict,
    LoraConfig,
    PeftType,
    PrefixTuningConfig,
    PromptEncoderConfig,
)

import evaluate
from datasets import load_dataset
from transformers import AutoModelForSequenceClassification, AutoTokenizer, get_linear_schedule_with_warmup, set_seed
from tqdm import tqdm
from accelerate import Accelerator, DistributedType
from accelerate.utils import set_seed
from accelerate import notebook_launcher


# In[5]:


def training_loop(mixed_precision="fp16", seed:int=42, batch_size:int=32):
    set_seed(seed)
    
    model_name_or_path = "google/flan-t5-small"
    task = "mrpc"
    
    # Initialize accelerator
    accelerator = Accelerator(mixed_precision=mixed_precision)
    
    if any(k in model_name_or_path for k in ("gpt", "opt", "bloom")):
        padding_side = "left"
    else:
        padding_side = "right"

    def collate_fn(examples):
        return tokenizer.pad(examples, padding="longest", return_tensors="pt")

    def collate_fn(examples):
        # On TPU it's best to pad everything to the same length or training will be very slow.
        max_length = 128 if accelerator.distributed_type == DistributedType.TPU else None
        # When using mixed precision we want round multiples of 8/16
        if accelerator.mixed_precision == "fp8":
            pad_to_multiple_of = 16
        elif accelerator.mixed_precision != "no":
            pad_to_multiple_of = 8
        else:
            pad_to_multiple_of = None

        return tokenizer.pad(
            examples,
            padding="longest",
            max_length=max_length,
            pad_to_multiple_of=pad_to_multiple_of,
            return_tensors="pt",
        )        
        
    def tokenize_function(examples):
        # max_length=None => use the model max length (it's actually the default)
        outputs = tokenizer(examples["sentence1"], examples["sentence2"], truncation=True, max_length=None)
        return outputs

    tokenizer = AutoTokenizer.from_pretrained(model_name_or_path, padding_side=padding_side)
    
    if getattr(tokenizer, "pad_token_id") is None:
        tokenizer.pad_token_id = tokenizer.eos_token_id

    datasets = load_dataset("glue", task)
    metric = evaluate.load("glue", task)

    tokenized_datasets = datasets.map(
        tokenize_function,
        batched=True,
        remove_columns=["idx", "sentence1", "sentence2"],
    )

    # We rename 'label' column to 'labels', which is expected for labels by models of the transformers library
    tokenized_datasets = tokenized_datasets.rename_column("label", "labels")
    
    # Build & Instantiate dataloaders.
    train_dataloader = DataLoader(
        tokenized_datasets["train"], 
        shuffle=True, 
        collate_fn=collate_fn, 
        batch_size=batch_size)
    eval_dataloader = DataLoader(
        tokenized_datasets["validation"], 
        shuffle=False, 
        collate_fn=collate_fn, 
        batch_size=batch_size)
    
    
    
    peft_type = PeftType.LORA
    num_epochs = 5
    
    peft_config = LoraConfig(
        task_type="SEQ_CLS", 
        inference_mode=False, 
        r=8, lora_alpha=16, 
        lora_dropout=0.1)
    lr = 3e-4
    
    model = AutoModelForSequenceClassification.from_pretrained(
        model_name_or_path, 
        return_dict=True)
    model = model.to(accelerator.device)
    model = get_peft_model(model, peft_config)
    
    optimizer = AdamW(params=model.parameters(), lr=lr)
    
    lr_scheduler = get_linear_schedule_with_warmup(
        optimizer=optimizer,
        num_warmup_steps=0.06 * (len(train_dataloader) * num_epochs),
        num_training_steps=(len(train_dataloader) * num_epochs),
    )    
    
    model, optimizer, train_dataloader, eval_dataloader, lr_scheduler = accelerator.prepare(
        model, optimizer, train_dataloader, eval_dataloader, lr_scheduler
    )


    for epoch in range(num_epochs):
        model.train()
        for step, batch in enumerate(tqdm(train_dataloader)):
            batch.to(accelerator.device)
            outputs = model(**batch)
            loss = outputs.loss
            accelerator.backward(loss)
            optimizer.step()
            lr_scheduler.step()
            optimizer.zero_grad()

        model.eval()
        for step, batch in enumerate(tqdm(eval_dataloader)):
            batch.to(accelerator.device)
            with torch.no_grad():
                outputs = model(**batch)
            predictions = outputs.logits.argmax(dim=-1)
            predictions, references = predictions, batch["labels"]
            metric.add_batch(
                predictions=predictions,
                references=references,
            )


# In[6]:


args = ("fp16", 42, 64)
notebook_launcher(training_loop, args, num_processes=2)


# In[ ]:




